// -*- C++ -*-
//
// This file is part of the Coriolis Software.
// Copyright (c) UPMC/LIP6 2008-2018, All Rights Reserved
//
// +-----------------------------------------------------------------+ 
// |                   C O R I O L I S                               |
// |     V L S I   B a c k e n d   D a t a - B a s e                 |
// |                                                                 |
// |  Author      :                    Jean-Paul CHAPUT              |
// |  E-mail      :       Jean-Paul.Chaput@asim.lip6.fr              |
// | =============================================================== |
// |  C++ Module  :       "./BreakpointWidget.cpp"                   |
// +-----------------------------------------------------------------+


#include <QSettings>
#include <QPointer>
#include <QApplication>
#include <QLabel>
#include <QPushButton>
#include <QSpinBox>
#include <QGridLayout>
#include <QFrame>
#include "hurricane/Breakpoint.h"
#include "hurricane/viewer/BreakpointWidget.h"
#include "hurricane/viewer/Graphics.h"


namespace Hurricane {


  QByteArray  BreakpointWidget::_geometry;


  void  BreakpointWidget::readQtSettings ()
  {
    QSettings settings;
    QString   sizeKey = QString( "BreakpointWidget/geometry" );
    if (not settings.contains(sizeKey)) return;
    settings.beginGroup( QString("BreakpointWidget") );
    _geometry = settings.value( "geometry" ).toByteArray();
    settings.endGroup();
  }


  void  BreakpointWidget::saveQtSettings ()
  {
    if (_geometry.isNull()) return;
    QSettings settings;
    settings.beginGroup( QString("BreakpointWidget") );
    settings.setValue( "geometry", _geometry );
    settings.endGroup();
  }


  BreakpointWidget::BreakpointWidget ( QWidget* parent )
    : QDialog    (parent)
    , _message   (new QLabel())
    , _stopLevel (new QSpinBox())
    , _isFinished(false)
    , _eventLoop (NULL)
  {
    setModal         ( false );
  //setWindowModality( Qt::WindowModal );
    setWindowTitle   ( "Breakpoint" );
    setToolTip       ( "Hit Chuck to continue..." );

  //_message->setTextFormat ( Qt::RichText );
    _message->setText       ( "<b>No Message Yet</b>" );
    _message->setSizePolicy ( QSizePolicy::Expanding, QSizePolicy::Expanding );

    QLabel* stopLabel = new QLabel ();
    stopLabel->setText ( "Adjust Stop Level:" );

    QPushButton* ok = new QPushButton ();
    ok->setIcon       ( QIcon(":/images/angry-birds-chuck.png") );
    ok->setFlat       ( true );
    ok->setIconSize   ( (Graphics::isHighDpi() ? QSize(200,200) : QSize(80,80)) );
    ok->setSizePolicy ( QSizePolicy::Fixed, QSizePolicy::Fixed );

    QFrame* vLine = new QFrame ();
    vLine->setFrameShape  ( QFrame::VLine );
    vLine->setFrameShadow ( QFrame::Sunken );

    QGridLayout* layout = new QGridLayout ();
  //layout->setSizeConstraint  ( QLayout::SetFixedSize );
    layout->addWidget          ( _message  , 0, 0, 1, 2 );
    layout->addWidget          ( stopLabel , 1, 0 );
    layout->addWidget          ( _stopLevel, 1, 1 );
    layout->addWidget          ( vLine     , 0, 2, 2, 1 );
    layout->addWidget          ( ok        , 0, 3, 2, 1, Qt::AlignVCenter );
    setLayout ( layout );

    connect ( ok        , SIGNAL(clicked())        , this, SLOT(accept()) );
    connect ( _stopLevel, SIGNAL(valueChanged(int)), this, SLOT(updateStopLevel(int)) );
    connect ( this      , SIGNAL(finished(int))    , this, SLOT(raiseFinished(int)) );
  }


  int  BreakpointWidget::execNoModal ()
  {
    if (isVisible()) return -1;

    // while (QApplication::hasPendingEvents()) {
    //   cerr << "Flush events from the main event loop." << endl;
    //   QApplication::processEvents();
    // }
    QApplication::flush();

    _isFinished = false;
    show ();
    if (not _geometry.isNull()) restoreGeometry( _geometry );

  // Snipet code from Qt's QDialog.
    _eventLoop = new QEventLoop ();
    QPointer<QDialog> guard = this;
    _eventLoop->exec( QEventLoop::DialogExec );
    _eventLoop = NULL;

    _geometry = saveGeometry();
    if (guard.isNull()) return QDialog::Rejected;

    return result();
  }


  void  BreakpointWidget::raiseFinished ( int )
  {
    _isFinished = true;
    if ( _eventLoop ) _eventLoop->exit(0);
  }


  void  BreakpointWidget::setMessage ( const QString& message )
  {
    _message->setText ( message );
  }


  void  BreakpointWidget::setStopLevel ( int level )
  {
    _stopLevel->setValue ( level );
  }


  int  BreakpointWidget::getStopLevel () const
  {
    return _stopLevel->value ();
  }


  void  BreakpointWidget::updateStopLevel ( int value )
  {
    unsigned int stopLevel = (value<0) ? 0 : value;
    Breakpoint::setStopLevel ( stopLevel );
  }


}  // End of Hurricane Namespace.
